// SPDX-FileCopyrightText: 2025 Tobias Leupold <tl@stonemx.de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes
#include "Style.h"

// Qt includes
#include <QDebug>
#include <QPainter>
#include <QStyleOption>
#include <QSplitter>
#include <QApplication>
#include <QFontMetrics>
#include <QLineEdit>
#include <QToolButton>

// C++ includes
#include <cmath>

Style::Style(QStyle *style)
    : QProxyStyle(style)
{
}

QSize Style::sizeFromContents(ContentsType type, const QStyleOption *option, const QSize &size,
                              const QWidget *widget) const
{
    if (type == CT_Splitter) {
        const auto *splitter = qobject_cast<const QSplitter *>(widget);
        if (splitter) {
            auto handleSize = QProxyStyle::sizeFromContents(type, option, size, widget);
            if (splitter->orientation() == Qt::Horizontal) {
                handleSize.setWidth(6);
            } else {
                handleSize.setHeight(6);
            }
            return handleSize;
        }

    }

    // Fallback
    return QProxyStyle::sizeFromContents(type, option, size, widget);
}

void Style::drawControl(ControlElement element, const QStyleOption *option, QPainter *painter,
                        const QWidget *widget) const
{
    // Draw a custom splitter handle

    // We skip those with nullptr parents to not catch the splitters used to display QDockWidgets,
    // only those inside them.

    if (element == CE_Splitter && widget->parent() != nullptr) {
        if (widget->property("notNeeded").toBool()) {
            return;
        }

        painter->save();

        QPalette palette;
        painter->setPen(Qt::NoPen);
        painter->setBrush(qApp->property("DARK_MODE").toBool() ? palette.color(QPalette::Light)
                                                               : palette.color(QPalette::Mid));

        auto rect = option->rect;
        if (rect.height() > rect.width()) {
            const auto padding = rect.height() * 0.2;
            rect.adjust(2, 2 + padding, -2, -2 - padding);
            painter->drawRect(rect);
        } else {
            const auto padding = rect.width() * 0.2;
            rect.adjust(2 + padding, 2, -2 - padding, -2);
            painter->drawRect(rect);
        }

        painter->restore();
        return;

    // Draw custom HLine/VLine QFrames

    } else if (element == CE_ShapedFrame) {
        const auto *frame = qobject_cast<const QFrame *>(widget);
        if (frame
            && (frame->frameShape() == QFrame::VLine || frame->frameShape() == QFrame::HLine)) {

            painter->save();

            QPalette palette;
            painter->setPen(palette.color(QPalette::Mid));

            if (frame->frameShape() == QFrame::VLine) {
                painter->drawLine(option->rect.topLeft(), option->rect.bottomLeft());
            } else {
                painter->drawLine(option->rect.topLeft(), option->rect.topRight());
            }

            painter->restore();
            return;
        }

    // Draw custom dock widget titles

    } else if (element == CE_DockWidgetTitle) {
        const auto *dockWidgetOption = qstyleoption_cast<const QStyleOptionDockWidget *>(option);
        if (dockWidgetOption) {
            // Copy the style option
            QStyleOptionDockWidget opt = *dockWidgetOption;

            // Set a bold font
            QFont font;
            font.setBold(true);
            font.setPointSize(font.pointSize() * 1.2);
            QFontMetrics fm(font);
            opt.fontMetrics = fm;
            painter->setFont(font);

            // Let the base style draw the title
            QProxyStyle::drawControl(element, &opt, painter, widget);
            return;
        }
    }

    // Fallback
    QProxyStyle::drawControl(element, option, painter, widget);
}

void Style::drawComplexControl(ComplexControl control, const QStyleOptionComplex *option,
                               QPainter *painter, const QWidget *widget) const
{
    // Draw bold group box titles if requested

    if (control == CC_GroupBox) {
        const auto *gbOption = qstyleoption_cast<const QStyleOptionGroupBox *>(option);
        if (! gbOption) {
            QProxyStyle::drawComplexControl(control, option, painter, widget);
            return;
        }

        const auto hasTitle = ! gbOption->text.isEmpty();
        const auto hasCheckBox = gbOption->subControls & SC_GroupBoxCheckBox;

        auto frRect = subControlRect(CC_GroupBox, gbOption, SC_GroupBoxFrame, widget);
        auto textRect = subControlRect(CC_GroupBox, gbOption, SC_GroupBoxLabel, widget);
        auto boxRect = subControlRect(CC_GroupBox, gbOption, SC_GroupBoxCheckBox, widget);

#ifdef Q_OS_MACOS
        if (hasTitle) {
            // MacOS uses a smaller font for the group box titles by default,
            // so we have to change the geometry a bit.
            QFont font;
            QFontMetrics fm(font);
            const auto vDiff = fm.height() - textRect.height() + 1;
            const auto hDiff = fm.boundingRect(gbOption->text).width() - textRect.width() + 1;
            textRect.adjust(0, 0, hDiff, vDiff);
            frRect.adjust(0, vDiff, 0, 0);
            boxRect.adjust(0, -(vDiff / 2), 0, 0);
        }
#endif

        // Draw the frame
        QStyleOptionFrame frOption;
        frOption.QStyleOption::operator=(*gbOption);
        frOption.features = gbOption->features;
        frOption.lineWidth = gbOption->lineWidth;
        frOption.midLineWidth = gbOption->midLineWidth;
        frOption.rect = frRect;
        drawPrimitive(PE_FrameGroupBox, &frOption, painter, widget);

        // Draw the title

        if ((gbOption->subControls & QStyle::SC_GroupBoxLabel) && hasTitle) {
            painter->save();

            if (! hasCheckBox && ! widget->property("normalTitle").toBool()) {
                // We want a bold title

                // Inform the painter
                QFont font;
                font.setBold(true);
                painter->setFont(font);
                QFontMetrics fm(font);

                // Adjust the label geometry

                const auto leftSpace = textRect.left();
                const auto rightSpace = frRect.width() - textRect.left() - textRect.width();
                if (std::abs(leftSpace - rightSpace) > 2) {
                    // Left-aligned
                    textRect.setWidth(fm.boundingRect(gbOption->text).width() + 2);
                } else {
                    // Centered
                    const auto width = fm.boundingRect(gbOption->text).width() + 2;
                    const auto diff = width - textRect.width();
                    textRect.setWidth(width);
                    textRect.translate(-diff / 2, 0);
                }
            }

#ifdef Q_OS_WINDOWS
            // On Windows, we first have to fill the background
            painter->fillRect(textRect, widget->palette().color(QPalette::Window));
#endif
            drawItemText(painter,
                         textRect,
                         gbOption->textAlignment | Qt::TextShowMnemonic,
                         gbOption->palette,
                         gbOption->state & State_Enabled,
                         gbOption->text,
                         QPalette::WindowText);

            painter->restore();
        }

        // Draw checkbox
        if (hasCheckBox) {
            QStyleOptionButton boxOpt;
            boxOpt.QStyleOption::operator=(*option);
            boxOpt.rect = boxRect;
            drawPrimitive(PE_IndicatorCheckBox, &boxOpt, painter, widget);
        }

        return;
    }

    QProxyStyle::drawComplexControl(control, option, painter, widget);
}
