// SPDX-FileCopyrightText: 2020-2025 Tobias Leupold <tl@stonemx.de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes
#include "Led.h"

// Qt includes
#include <QTimer>
#include <QFontMetrics>
#include <QDebug>
#include <QPen>
#include <QPainter>
#include <QApplication>

// C++ includes
#include <functional>

static const QBrush s_offLight(QColor(220, 220, 220));
static const QBrush s_offDark (QColor( 90,  90,  90));
static const QBrush s_red     (QColor(230,  10,  10));
static const QBrush s_green   (QColor( 10, 230,  10));

Led::Led(int size, QWidget *parent) : QLabel(parent)
{
    // Make the size even
    size = size * 2 / 2;

    // Cache the size and radius
    m_rect = QRect(0, 0, size, size);
    m_radius = size / 2;
    setFixedSize(size, size);

    // Setup the flash timer
    m_flashTimer = new QTimer(this);
    m_flashTimer->setInterval(250);
    m_flashTimer->setSingleShot(true);
    connect(m_flashTimer, &QTimer::timeout,
            this, std::bind(&Led::setState, this, State::Off));
}

void Led::paintEvent(QPaintEvent *)
{
    QPainter painter(this);
    painter.setRenderHint(QPainter::Antialiasing, true);

    const auto darkMode = qApp->property("DARK_MODE").toBool();

    // Border
    painter.setPen(QPen(darkMode ? palette().light().color()
                                 : palette().mid().color()));

    // Color
    switch (m_state) {
    case Led::Off:
        painter.setBrush(darkMode ? s_offDark
                                  : s_offLight);
        break;
    case Led::Red:
        painter.setBrush(s_red);
        break;
    case Led::Green:
        painter.setBrush(s_green);
        break;
    }

    painter.drawRoundedRect(m_rect, m_radius, m_radius);
}

void Led::setState(State state)
{
    m_state = state;
    repaint();
}

void Led::flash()
{
    setState(Led::Green);
    m_flashTimer->start();
}
