// SPDX-FileCopyrightText: 2025 Tobias Leupold <tl@stonemx.de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes
#include "PopupButton.h"

// Qt includes
#include <QDebug>
#include <QStyleOptionButton>
#include <QPainter>
#include <QPainterPath>
#include <QApplication>

static const int s_margin = 1;
static const int s_radius = 5;

PopupButton::PopupButton(const QString &text, QWidget *parent)
    : QPushButton(text, parent)
{
#ifdef Q_OS_MACOS
    // Needed for the hover indicator to work on macOS
    setAttribute(Qt::WA_Hover, true);
#endif
}

void PopupButton::setColor(const QColor &color)
{
    m_color = color;
}

void PopupButton::paintEvent(QPaintEvent *)
{
    QStyleOptionButton option;
    option.initFrom(this);

    QPainter painter(this);
    painter.setRenderHint(QPainter::Antialiasing);

    // Draw a border if the mouse hovers the button
    // This is adapted from KDE's Breeze's drawFocusFrame()

    if (option.state & QStyle::State_MouseOver) {
        const auto inner = rect().adjusted(s_margin, s_margin, -s_margin, -s_margin);

        QPainterPath painterPath;
        painterPath.addRoundedRect(inner, s_radius, s_radius);

        // Indicate a click
        if (isDown()) {
            painter.fillPath(
                painterPath,
#ifndef Q_OS_MACOS
                qApp->property("DARK_MODE").toBool()
                    ? palette().color(QPalette::Highlight).darker(330)
                    : palette().color(QPalette::Highlight).lighter(180)
#else
                // On macOS, the "Hightlight" color is already quite light
                palette().color(QPalette::Highlight)
#endif
            );
        }

        // Draw the border
        painterPath.setFillRule(Qt::OddEvenFill);
        painterPath.addRoundedRect(rect(), s_radius + s_margin, s_radius + s_margin);
        painter.fillPath(painterPath, palette().color(QPalette::Highlight));
    }

    // Draw the button's text
    if (m_color.isValid()) {
        painter.setPen(m_color);
    }
    painter.drawText(rect(), Qt::AlignCenter, text());

    // Draw the button's icon if one is set
    if (! icon().isNull()) {
        const auto pixmap = icon().pixmap(iconSize(), isEnabled() ? QIcon::Normal
                                                                  : QIcon::Disabled);
        painter.drawPixmap(QPoint((width() - pixmap.width()) / 2, (height() - pixmap.height()) / 2),
                           pixmap);
    }
}
