// SPDX-FileCopyrightText: 2019-2025 Tobias Leupold <tl@stonemx.de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes
#include "SelectionPopup.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QCursor>
#include <QScrollArea>
#include <QScrollBar>
#include <QApplication>
#include <QTimer>
#include <QScreen>
#include <QPainter>

SelectionPopup::SelectionPopup(QWidget *parent) : QDialog(parent)
{
    setWindowFlags(Qt::FramelessWindowHint | Qt::Popup);

    auto *layout = new QVBoxLayout(this);
    layout->setSpacing(0);
    layout->setContentsMargins(1, 1, 1, 1); // We need this for the QDialog's gray border

    m_scrollArea = new QScrollArea(this);
    m_scrollArea->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    m_scrollArea->setFrameShape(QFrame::NoFrame);
    m_scrollArea->viewport()->setAutoFillBackground(false);
    layout->addWidget(m_scrollArea);
}

void SelectionPopup::paintEvent(QPaintEvent *)
{
    QPainter painter(this);
    const QPalette palette;
    painter.fillRect(QRect(0, 0, width(), height()), palette.color(QPalette::Base));
    painter.setPen(qApp->property("DARK_MODE").toBool() ? palette.color(QPalette::Light)
                                                        : palette.color(QPalette::Mid));
    painter.drawRect(QRect(0, 0, width() - 1, height() - 1));
}

void SelectionPopup::setScrollWidget(QWidget *widget)
{
    m_scrollArea->setWidget(widget);
    widget->setAutoFillBackground(false);
}

void SelectionPopup::show(QPoint position)
{
    Q_EMIT aboutToShow();

    if (position == QPoint(-1, -1)) {
        position = QCursor::pos();
    }

    const QRect availableGeometry = QGuiApplication::primaryScreen()->availableGeometry();
    const int availableHeight = availableGeometry.height() - position.y();
    const int availableWidth = availableGeometry.width() - position.x();

    const QSize widgetSize = m_scrollArea->widget()->size();
    const int widgetHeight = widgetSize.height() + 2; // Two pixels for the border
    const int widgetWidth = widgetSize.width(); // Who appararenly aren't necessary here

    if (availableWidth < widgetWidth) {
        if (availableGeometry.width() >= widgetWidth) {
            position.setX(position.x() - (widgetWidth - availableWidth));
        } else {
            position.setX(0);
        }
    }

    int targetHeight = -1;

    if (widgetHeight <= availableHeight) {
        targetHeight = widgetHeight;
    } else {
        if (availableGeometry.height() >= widgetHeight) {
            position.setY(position.y() - (widgetHeight - availableHeight));
            targetHeight = widgetHeight;
        } else {
            position.setY(0);
            targetHeight = availableGeometry.height();
        }
    }

    resize(m_scrollArea->widget()->width(), targetHeight);
    move(position);

    QDialog::show();
    QTimer::singleShot(0, this, [this]
    {
        if (m_scrollArea->verticalScrollBar()->isVisible()) {
            resize(m_scrollArea->widget()->width() + m_scrollArea->verticalScrollBar()->width(),
                height());
        } else {
            resize(m_scrollArea->widget()->width(), height());
        }
    });
}
